perl -I../lib check-run.t 1
perl -I../lib check-script=run.t 1
perl -I../lib check_fail-run.t 1
perl -I../lib check_fail-script=run.t 1
perl -I../lib check_retry-run.t 1
perl -I../lib check_retry-script=run.2.t 1
perl -I../lib check_retry-script=run.t 1
perl -I../lib error.t 1
perl -I../lib fail-run.t 1
perl -I../lib fail-run_continue.t 1
perl -I../lib fail-run_display.t 1
perl -I../lib fail-script=run.t 1
perl -I../lib for-run.t 1
perl -I../lib for-script.t 1
perl -I../lib for-script=script.t 1
perl -I../lib for_fail-run.t 1
perl -I../lib for_fail-run_continue.t 1
perl -I../lib for_fail-script.t 1
perl -I../lib for_fail-script=script.t 1
perl -I../lib if-run.t 1
perl -I../lib if-script=run.t 1
perl -I../lib obj.t 1
perl -I../lib retry-run.t 1
perl -I../lib retry-script=run.t 1
perl -I../lib retry_fail-run.t 1
perl -I../lib retry_fail-run_continue.t 1
perl -I../lib simple-run.t 1
perl -I../lib simple-run_echo.t 1
perl -I../lib simple-run_output=merged.t 1
perl -I../lib simple-run_output=quiet.t 1
perl -I../lib simple-run_output=stderr.t 1
perl -I../lib simple-run_output=stdout.t 1
perl -I../lib simple-script.t 1
perl -I../lib simple-script=run.t 1
perl -I../lib simple-script=run_output=stdout.t 1
perl -I../lib simple-script=script.t 1
perl -I../lib simple-script=simple.t 1
perl -I../lib special_characters-run.t 1
perl -I../lib special_characters-script=run.t 1
perl -I../lib until-run.t 1
perl -I../lib until-script.t 1
perl -I../lib until-script=script.t 1
perl -I../lib while-run.t 1
perl -I../lib while-script.t 1
perl -I../lib while-script=script.t 1
perl -I../lib zz-cleanup.t 1
